"""repository abstraction for a mercurial repository

:organization: Logilab
:copyright: 2007-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

from __future__ import with_statement

__docformat__ = "restructuredtext en"

from os import umask, makedirs
from os.path import split, exists, join, dirname
from datetime import datetime

from mercurial import hg, ui, revlog, context, util, commands as hgcmd
from mercurial.node import nullid, short as short_hex

from cubicweb import Binary, QueryError
from cubicweb.server.ssplanner import EditedEntity

from cubes.vcsfile import repo, bridge
from cubes.vcsfile.entities import remove_authinfo

_DELETED = object()

def monkeypatch_mercurial_url():
    from mercurial.url import httphandler
    if hasattr(httphandler, '__del__'):
        del httphandler.__del__
monkeypatch_mercurial_url()


class vcsfile_ui(ui.ui):
    """ redirect output to cubicweb log """
    def __init__(self, loggable):
        ui.ui.__init__(self)
        self.cw_loggable = loggable

    def copy(self):
        return self

    def interactive(self):
        return False

    # XXX beware encoding issues wrt localized outputs ...
    def write(self, *args, **kwargs):
        if self._buffers:
            self._buffers[-1].extend([str(a) for a in args])
        else:
            for a in args:
                self.cw_loggable.debug(str(a).strip())

    def write_err(self, *args, **kwargs):
        for a in args:
            self.cw_loggable.error(str(a).strip())


def bw_nb_revisions(changelog):
    """return the number of revisions in a repository's changelog,
    assuring mercurial < 1.1 compatibility
    """
    try:
        return len(changelog)
    except AttributeError: # hg < 1.1
        return changelog.count()


class HGRepository(repo.VCSRepository):
    type = 'hg'

    def pull_or_clone_cache(self, url):
        if not exists(self.path):
            pdir = dirname(self.path)
            if not exists(pdir):
                makedirs(pdir)
            self.info('cloning %s into %s', url, self.path)
            hgcmd.clone(vcsfile_ui(self), str(url),
                        dest=self.path, stream='compressed', noupdate=True)
        else:
            self.info('pulling from %s into %s', remove_authinfo(url), self.path)
            vui = vcsfile_ui(self)
            hgcmd.pull(vui, hg.repository(vui, self.path), source=str(url))

    def import_content(self, repoentity, commitevery=10):
        """import content from the mercurial repository"""
        assert self.eid == repoentity.eid
        if not exists(self.path):
            self.info('repository %s not accessible in %s',
                      repoentity.dc_title(), self.path)
            return
        try:
            repo = self.hgrepo()
        except Exception, ex:
            self.exception('Impossible to open repository %s (%s)', self, ex)
            return
        # actual rev is the *number* of revision, latest rev the highest
        # revision number while hg revisions start at 0 -> - 1
        actualrev = bw_nb_revisions(repo.changelog) - 1
        session = repoentity._cw
        latestrev = repoentity.latest_known_revision()
        if actualrev < latestrev:
            # striped repo
            # very minimal support, this won't work if local number of
            # revisions has changed due to the strip
            self.warning('%s: strip detected ? latest known revision is %s, '
                         'latest repository revision is %s', self.path,
                         latestrev, actualrev)
            session.execute('DELETE Revision X WHERE X revision > %(rev)s',
                            {'rev': actualrev})
            return
        self.info('hg repo %s: known revisions up to %s, repo revision %s',
                  self.path, latestrev, actualrev)
        for i in xrange(latestrev + 1, actualrev + 1):
            self.info('importing revision %s', i)
            try:
                self.import_revision(repoentity, repo, i)
            except:
                self.critical('error while importing revision %s of %s',
                              i, self.path, exc_info=True)
                raise
            if not i % commitevery:
                # give a change to threads waiting for a pool by freeing it then
                # reacquiring it
                session.commit()
                session.set_pool()

    def import_revision(self, repoentity, repo, i):
        session = repoentity._cw
        execute = session.execute
        node = repo.changelog.node(i)
        changeset = unicode(short_hex(node))
        if execute('Any X WHERE X from_repository R, R eid %(repo)s, X changeset %(cs)s',
                   {'repo': repoentity.eid, 'cs': changeset}):
            self.warning('skip revision %s, seems already imported',
                         changeset)
            return
        ctx = repo.changectx(i)
        date = datetime.fromtimestamp(ctx.date()[0])
        #taglist = ctx.tags() #repo.nodetags(node)
        revdata = {'date': date, 'revision': i,
                   'author': unicode(ctx.user(), self.encoding, 'replace'),
                   'description': unicode(ctx.description(), self.encoding, 'replace'),
                   'changeset': changeset,
                   'branch': unicode(ctx.branch()),
                   }
        self.imported_revision = i # for hooks
        parents = [short_hex(n) for n in repo.changelog.parents(node)
                   if n != nullid]
        if parents:
            if len(parents) == 1:
                parents = execute('Revision X WHERE X changeset %(cs)s, '
                                  'X from_repository R, R eid %(r)s',
                                  {'cs': parents[0], 'r': self.eid})
            else:
                parents = execute('Revision X WHERE X changeset IN (%s), '
                                  'X from_repository R, R eid %%(r)s'
                                  % ','.join("'%s'"%cs for cs in parents),
                                  {'r': self.eid})
        revdata['parents'] = [r[0] for r in parents]
        reveid = bridge.import_revision(session, self.eid, **revdata)
        if not repoentity._cw.vreg.config['import-revision-content']:
            return
        changes = repo.status(ctx.parents()[0].node(), ctx.node())[:3]
        modified, added, removed = changes
        path_filter = repoentity.subpath
        for path in modified + added:
            upath = unicode(path, self.encoding, 'replace')
            if not (path_filter is None or upath.startswith(path_filter)):
                continue
            directory, fname = split(path)
            session.transaction_data['vcs_importing'] = self.eid, directory, fname, i
            filectx = ctx[path]
            vcdata = {'data': Binary(filectx.data())}
            renameinfo = filectx.renamed()
            if renameinfo:
                oldfile, fileid = renameinfo
                pfctx = repo.filectx(oldfile, fileid=fileid)
                pcs = short_hex(pfctx.node())
                dir, name = split(unicode(oldfile, self.encoding, 'replace'))
                pvc = execute('VersionContent X WHERE '
                              'X from_revision REV, REV changeset %(cs)s, '
                              'REV from_repository R, R eid %(r)s, '
                              'X content_for VF, VF directory %(dir)s, VF name %(name)s',
                              {'cs':  pcs, 'r': self.eid,
                               'dir': dir, 'name': name})
                if pvc:
                    assert len(pvc) == 1
                    if oldfile in removed:
                        vcdata['vc_rename'] = pvc[0][0]
                    else:
                        vcdata['vc_copy'] = pvc[0][0]
                else:
                    self.error('detected copy or rename of %s@%s but unable'
                               ' to find associated version content',
                               oldfile, pcs)
            bridge.import_version_content(session, self.eid, reveid, upath,
                                          date, **vcdata)
        for path in removed:
            upath = unicode(path, self.encoding, 'replace')
            if not (path_filter is None or upath.startswith(path_filter)):
                continue
            bridge.import_deleted_version_content(session, self.eid, reveid,
                                                  upath, date)
        bridge.set_at_revision(session, reveid)

    def hgrepo(self):
        return hg.repository(ui.ui(), self.path)

    def _file_content(self, path, rev):
        """extract a binary string with the content of the file at `path` for
        revision `rev` in the repository
        """
        ctx = self.hgrepo().changectx(rev)
        return Binary(ctx[path].data())

    def revision_transaction(self, session, entity):
        """open a transaction to create a new revision corresponding to the
        given entity
        """
        return HGTransaction(self, session, entity)

    def add_versioned_file_content(self, session, transaction, vf, entity,
                                   data):
        """add a new revision of a versioned file"""
        vfpath = self.encode(vf.path)
        transaction.changes[vfpath] = context.memfilectx(
            vfpath, data.getvalue(), islink=False, isexec=False, copied=None)
        return vf.path

    def add_versioned_file_deleted_content(self, session, transaction, vf,
                                           entity):
        """add a new revision of a just deleted versioned file"""
        transaction.changes[self.encode(vf.path)] = _DELETED
        return vf.path


class HGTransaction(object):
    def __init__(self, repohdlr, session, revision):
        self.repohdlr = repohdlr
        self.session = session
        self.revision = revision
        self.repo = repohdlr.hgrepo()
        self._wlock = self.repo.wlock()
        self._lock = self.repo.lock()
        self.extra = {}
        self.changes = {}

    @property
    def rev(self):
        """newly created revision number"""
        return bw_nb_revisions(self.repo.changelog)

    def _filectx(self, repo, memctx, path):
        """callable receiving the repository, the current memctx object and the
        normalized path of requested file, relative to repository root. It is
        fired by the commit function for every file in 'files', but calls order
        is undefined. If the file is available in the revision being committed
        (updated or added), filectxfn returns a memfilectx object. If the file
        was removed, filectxfn raises an IOError. Moved files are represented by
        marking the source file removed and the new file added with copy
        information (see memfilectx).
        """
        if self.changes[path] is _DELETED:
            raise IOError()
        return self.changes[path]

    def precommit(self):
        if not self.changes:
            raise QueryError('nothing changed')

    def commit(self):
        # XXX: make umask configurable
        oldumask = umask(022)
        try:
            self._commit()
        finally:
            umask(oldumask)

    def _commit(self):
        # XXX merging branches
        encode = self.repohdlr.encode
        branch = encode(self.revision.cw_attr_cache.get('branch') or u'default')
        if branch != u'default':
            self.extra['branch'] = branch
        author = encode(self.revision.cw_attr_cache.get('author', u''))
        msg = encode(self.revision.cw_attr_cache.get('description', u''))
        try:
            p1 = self.repo.branchtags()[branch]
        except KeyError:
            # new branch
            p1 = self.repo.branchtags().get('default')
        p2 = None
        # ensure mercurial will use configured repo encoding, not locale's
        # encoding
        # XXX modifying module's global is not very nice but I've no other idea
        try:
            # mercurial >= 1.3.1
            from mercurial import encoding
            encoding.encoding = self.repohdlr.encoding
        except ImportError:
            util._encoding = self.repohdlr.encoding
        ctx = context.memctx(self.repo, (p1, p2), msg, self.changes.keys(),
                             self._filectx, author, extra=self.extra)
        try:
            # mercurial < 1.3.1
            node = self.repo._commitctx(ctx, force=True, force_editor=False,
                                        empty_ok=True, use_dirstate=False,
                                        update_dirstate=False)
        except AttributeError:
            node = self.repo.commitctx(ctx)
        # update revision's record to set correct changeset
        #
        # take care, self.repo != self.session.repo (eg mercurial repo instance
        # / rql repository)
        source = self.session.repo.system_source
        # remove previous cached value to only save the changeset
        with setattrcm(self.revision, 'cw_edited', EditedEntity(self.revision)) as rev:
            rev.cw_edited['changeset'] = unicode(short_hex(node))
            source.update_entity(self.session, rev)
        # XXX restore entity's dict?
        self.rollback()

    def rollback(self):
        try:
            # mercurial >= 1.3.1
            self._wlock.release()
            self._lock.release()
        except AttributeError:
            del self._wlock
            del self._lock



class setattrcm(object):
    def __init__(self, obj, attr, value):
        self.obj = obj
        self.attr = attr
        self.value = value

    def __enter__(self):
        self.old = getattr(self.obj, self.attr)
        setattr(self.obj, self.attr, self.value)
        return self.obj

    def __exit__(self, exctype, value, traceback):
        setattr(self.obj, self.attr, self.old)


from logging import getLogger
from cubicweb import set_log_methods
set_log_methods(HGRepository, getLogger('cubicweb.sources.hg'))



import os
from cubes.vcsfile.hooks import repo_cache_dir

sync_schema_props_perms(('Repository', 'title', 'String'), syncperms=False)


cacheroot = repo_cache_dir(repo.config)
for repo in rql('Any X, XLC WHERE X local_cache XLC, NOT X local_cache NULL,'
                'X is Repository').entities():
    if repo.local_cache.startswith(cacheroot):
        repo.set_attributes(local_cache=repo.local_cache.replace(cacheroot, '').lstrip(os.sep))




"""repository abstraction for a subversion repository

:organization: Logilab
:copyright: 2007-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from os.path import join, sep, split
from datetime import datetime
from cStringIO import StringIO

from logilab.common.decorators import cached

from svn import fs, repos, core, delta, ra

from cubicweb import Binary, QueryError

from cubes.vcsfile import repo, bridge

def _read_file_content(root, path):
    content = Binary()
    try:
        stream = fs.file_contents(root, path)
    except:
        return None
    while 1:
        data = core.svn_stream_read(stream, core.SVN_STREAM_CHUNK_SIZE)
        content.write(data)
        if len(data) < core.SVN_STREAM_CHUNK_SIZE:
            break
    core.svn_stream_close(stream)
    return content

class SVNRepository(repo.VCSRepository):
    type = 'svn'

    def __init__(self, *args, **kwargs):
        super(SVNRepository, self).__init__(*args, **kwargs)
        # svn uses utf-8 internally; we should not try to contradict that
        self.encoding = 'utf-8'

    def canonicalize(self, fspath):
        fspath = core.svn_path_canonicalize(fspath)
        repos.open(fspath) # throwaway, just check it's a valid svn repository
        return fspath

    def import_content(self, repoentity, commitevery=10):
        """import content from the subversion repository

        NOTE: server side only method
        """
        try:
            repo = repos.open(self.path)
            fsrep = repos.fs(repo)
        except core.SubversionException, ex:
            self.error('Impossible to open repository %s (%s)', self, ex)
            return
        callbacks = ra.Callbacks()
        #ra_ctx = ra.open2('file://' + self.path, callbacks, None, None)
        rev = repoentity.latest_known_revision()
        if rev == -1:
            rev = 0
        actualrev = fs.youngest_rev(fsrep)
        session = repoentity._cw
        while rev < actualrev:
            base_root = fs.revision_root(fsrep, rev)
            rev += 1
            if session.execute('Any X WHERE X from_repository R, R eid %(repo)s,'
                               'X revision %(rev)s',
                               {'repo': repoentity.eid, 'rev': rev}):
                self.warning('skip revision %s, seems already imported', rev)
                continue
            self.info('importing content of %s for revision %s', self, rev)
            self.imported_revision = rev # for hooks
            # the base of the comparison
            editor = SVNImporter(repoentity, fsrep, base_root, rev,
                                 import_revision_content = repoentity._cw
                                 .vreg.config['import-revision-content'])
            try:
                editor.import_revision()
            except:
                self.critical('error while importing revision %s of %s',
                              rev, repoentity.path, exc_info=True)
                raise
            if not rev % commitevery:
                # give a change to threads waiting for a pool by freeing it then
                # reacquiring it
                session.commit()
                session.set_pool()

    def _file_content(self, path, rev):
        """extract a binary string with the content of the file at `path` for
        revision `rev` in the repository
        """
        fsrep = repos.fs(repos.open(self.path))
        root = fs.revision_root(fsrep, rev)
        return _read_file_content(root, path)

    def revision_transaction(self, session, entity):
        """open a transaction to create a new revision corresponding to the
        given entity
        """
        if session.transaction_data.get('%s_info' % self.eid):
            repoptr, fsrep = session.transaction_data.get('%s_info' % self.eid)
        else:
            repoptr = repos.open(self.path)
            fsrep = repos.fs(repoptr)
            session.transaction_data['%s_info' % self.eid] = (repoptr, fsrep)
        # open a transaction against HEAD
        return SVNTransaction(repoptr, fsrep,
                              self.encode(entity.cw_attr_cache.get('author', u'')),
                              self.encode(entity.cw_attr_cache.get('description', u'')))

    def add_versioned_file_content(self, session, transaction, vf, entity, data):
        """add a new revision of a versioned file"""
        root = transaction.root
        # check if the parent directory exists in the repository
        directory = vf.directory.encode(self.encoding)
        if directory:
            parts = directory.split(sep)
            for i in xrange(len(parts)):
                directory = sep.join(parts[:i+1])
                kind = fs.check_path(root, directory)
                if kind == core.svn_node_none:
                    fs.make_dir(root, directory)
        # check if the file exists in the repository
        fname = join(directory, vf.name.encode(self.encoding))
        kind = fs.check_path(root, fname)
        if kind == core.svn_node_none:
            self.info('file %r does not exist in repo %s, creating...',
                      fname, self)
            fs.make_file(root, fname)
        else:
            self.info('updating file %r in repo %s', fname, self)
        handler, baton = fs.apply_textdelta(root, fname, None, None)
        delta.svn_txdelta_send_string(data.getvalue(), handler, baton)
        transaction.modified.add(fname)
        return fname.decode(self.encoding)

    def add_versioned_file_deleted_content(self, session, transaction, vf, entity):
        """add a new revision of a just deleted versioned file"""
        root = transaction.root
        # check if the file exists in the repository
        directory = vf.directory.encode(self.encoding)
        fname = join(directory, vf.name.encode(self.encoding))
        kind = fs.check_path(root, fname)
        if kind == core.svn_node_none:
            self.debug('file %r does not exist in repo %s',
                      fname, self)
            raise Exception('file %r does not exist' % fname)
        self.info('deleting %r from repo', fname)
        fs.delete(root, fname)
        transaction.modified.add(fname)
        return fname.decode(self.encoding)


class SVNTransaction(object):
    def __init__(self, repoptr, fsrep, author, msg):
        self.modified = set()
        rev = fs.youngest_rev(fsrep)
        txn = repos.fs_begin_txn_for_commit(repoptr, rev, author, msg)
        root = fs.txn_root(txn)
        #rev_root = fs.revision_root(fsrep, rev)
        self.root = root
        self.rev = rev+1
        self.txn = txn
        self.repoptr = repoptr

    def precommit(self):
        if not self.modified:
            raise QueryError('nothing changed')

    def commit(self):
        repos.fs_commit_txn(self.repoptr, self.txn)

    def rollback(self):
        fs.abort_txn(self.txn)


class SVNImporter(delta.Editor):
    def __init__(self, repoentity, fsrep, base_root, rev,
                 import_revision_content=True):
        self.repoeid = repoentity.eid
        self.encoding = repoentity.encoding
        self.session = repoentity._cw
        self.import_revision_content = import_revision_content
        path_filter = repoentity.subpath
        if isinstance(path_filter, unicode):
            # if we don't encode we may get unicode error when comparing files'
            # path against the path filter
            path_filter = path_filter.encode(self.encoding)
        self.path_filter = path_filter
        self.fsrep = fsrep
        #self.svnrepo = repo
        #self.ra_ctx = ra_ctx
        self.base_root = base_root
        self.rev = rev
        self.root = fs.revision_root(fsrep, rev)
        self.execute = self.session.execute
        revkwargs = self.base_rev_content()
        self.reveid = bridge.import_revision(self.session, self.repoeid,
                                             **revkwargs)
        self.date = revkwargs['date']

    def import_revision(self):
        def authz_cb(root, path, pool):
            return 1
        # construct the editor
        e_ptr, e_baton = delta.make_editor(self)
        # compute the delta
        repos.dir_delta(self.base_root, '', '', self.root, '',
                        e_ptr, e_baton, authz_cb, 0, 1, 0, 0)
        bridge.set_at_revision(self.session, self.reveid)

    def base_rev_content(self):
        revdata = {'revision': self.rev}
        author = fs.revision_prop(self.fsrep, self.rev,
                                  core.SVN_PROP_REVISION_AUTHOR)
        revdata['author'] = unicode(author, self.encoding, 'replace')
        rawdate = fs.revision_prop(self.fsrep, self.rev,
                                   core.SVN_PROP_REVISION_DATE)
        if rawdate:
            aprtime = core.svn_time_from_cstring(rawdate)
            # aprtime is microseconds; make seconds
            # assume secs in local TZ XXX we don't really know the TZ, do we?
            revdata['date'] = datetime.fromtimestamp(aprtime / 1000000)
        else:
            revdata['date'] = datetime.now()
        description = fs.revision_prop(self.fsrep, self.rev,
                                       core.SVN_PROP_REVISION_LOG)
        revdata['description'] = unicode(description, self.encoding,
                                       'replace')
        parent = self.execute('Revision X WHERE X revision %(rev)s, '
                              'X from_repository R, R eid %(r)s',
                              {'rev': self.rev - 1, 'r': self.repoeid})
        revdata['parents'] = parent and parent[0]
        return revdata

#     def rev_content(self, path):
#         # XXX code below may be used to get branches/tags information
#         # note mergeinfo require svn >= 1.5 (an update of the repo backend may
#         # be necessary)
##         print path, self.rev
##         print 'LOCS', ra.get_locations(self.ra_ctx, path, self.rev, range(1, self.rev))
##         print core.SVN_PROP_MERGEINFO, fs.node_prop(self.root, path, core.SVN_PROP_MERGEINFO)
##         try:
##             print 'FMI {',
##             minfos = repos.fs_get_mergeinfo(self.svnrepo, [path], self.rev,
##                                             core.svn_mergeinfo_inherited,
##                                             False, None, None)
##             for mf, mfinfos in minfos.iteritems():
##                 print mf, ': {',
##                 for mergedf, revranges in mfinfos.iteritems():
##                     print mergedf, [(rev.start, rev.end) for rev in revranges], ',',
##                 print '}',
##             print '}'
##         except core.SubversionException:
##             print 'not supported'
#         return fdata

    def match_path(self, path):
        return self.path_filter is None or path.startswith(self.path_filter)

    # svn generated events ####################################################

    def open_root(self, base_revision, dir_pool):
        #print 'open_root', base_revision, dir_pool
        pass

    def change_dir_prop(self, dir_baton, name, value, pool):
        #print 'change_dir_prop', dir_baton, name, value, pool
        pass

    def open_directory(self, path, *args):
        #print 'open_directory', path
        pass

    def close_directory(self, baton):
        #print 'close_directory', baton
        pass

    def _mark_importing(self, path):
        dir, name = split(path)
        self.session.transaction_data['vcs_importing'] = (self.repoeid, dir,
                                                          name, self.rev)
    def add_file(self, path, *args):
        if self .import_revision_content and self.match_path(path):
            self._mark_importing(path)
            upath = unicode(path, self.encoding, 'replace')
            bridge.import_version_content(
                self.session, self.repoeid, self.reveid, upath, self.date,
                data=_read_file_content(self.root, path))

    def add_directory(self, path, *args):
        #print 'add_directory', path
        pass

    def change_file_prop(self, file_baton, name, value, pool):
        #print 'change_file_prop', file_baton, name, value, pool
        pass

    def open_file(self, path, *args):
        #print 'open file', path
        if self.match_path(path):
            self._mark_importing(path)
            upath = unicode(path, self.encoding, 'replace')
            bridge.import_version_content(
                self.session, self.repoeid, self.reveid, upath, self.date,
                data=_read_file_content(self.root, path))

    def close_file(self, file_baton, text_checksum):
        pass

    def delete_entry(self, path, revision, parent_baton, pool):
        if self.match_path(path):
            # can't use svn property, return svn_core_none for deleted items
            upath = unicode(path, self.encoding, 'replace')
            directory, name = split(upath)
            if self.execute(
                'Any X WHERE X is VersionedFile, X directory %(path)s, '
                'X name %(name)s, X from_repository R, R eid %(r)s',
                {'path': directory, 'name': name, 'r': self.repoeid}):
                # deleted file
                bridge.import_deleted_version_content(
                    self.session, self.repoeid, self.reveid, upath, self.date)
                return
            # deleted directory
            for childdir, childname in self.execute(
                'Any XD, XN WHERE X is VersionedFile, X directory ~= %(path)s, '
                'X from_repository R, R eid %(r)s, X directory XD, X name XN',
                {'path': path + '%', 'r': self.repoeid}):
                upath = join(childdir, childname)
                bridge.import_deleted_version_content(
                    self.session, self.repoeid, self.reveid, upath, self.date)

    def apply_textdelta(self, file_baton, base_checksum):
        #print 'apply_textdelta', file_baton, base_checksum
        pass



"""vcsfile.views module tests"""

import re
from os import remove
from os.path import exists

from cubicweb.web.views import actions, primary, idownloadable

from utils import VCSFileTC, remove_eid # import first to setup import machinery
from cubes.vcsfile import entities
from cubes.vcsfile.views import primary as svnfviews, actions as vcsactions



class VCViewsTC(VCSFileTC):

    def test_entity(self):
        v1 = self.execute('Any X WHERE X from_revision R, R revision 1, X content_for VF, VF name "toto.txt"').get_entity(0, 0)
        self.assertIsInstance(v1, entities.VersionContent)
        self.assertEqual(v1.file.path, "toto.txt")
        self.assertEqual(v1.is_head(), False)
        self.assertEqual(v1.head.revision, 2)
        self.assertEqual(v1.cw_adapt_to('IPrevNext').previous_entity(), None)
        self.assertEqual(v1.cw_adapt_to('IPrevNext').next_entity().revision, 2)
        v2 = self.execute('Any X WHERE X from_revision R, R revision 2, X content_for VF, VF name "toto.txt"').get_entity(0, 0)
        self.assertIsInstance(v2, entities.VersionContent)
        self.assertEqual(v2.file.path, "toto.txt")
        self.assertEqual(v2.is_head(), True)
        self.assertEqual(v2.head.revision, 2)
        self.assertEqual(v2.cw_adapt_to('IPrevNext').previous_entity().revision, 1)
        self.assertEqual(v2.cw_adapt_to('IPrevNext').next_entity(), None)
        v3 = self.execute('Any X WHERE X from_revision R, R revision 3').get_entity(0, 0)
        self.assertIsInstance(v3, entities.DeletedVersionContent)
        self.failIf(isinstance(v3, entities.VersionContent)) # VersionContent(DeletedVersionContent)
        self.assertEqual(v3.file.path, "dir1/data.bin.gz")
        self.assertEqual(v3.head.revision, 3)
        self.assertEqual(v3.is_head(), True)
        self.assertEqual(v3.cw_adapt_to('IPrevNext').previous_entity().revision, 1)
        self.assertEqual(v3.cw_adapt_to('IPrevNext').next_entity(), None)

    def test_possible_actions(self):
        req = self.request()
        v1rset = req.execute('Any X WHERE X from_revision R, R revision 1, X content_for VF, VF name "toto.txt"')
        self.assertEqual(self.pactions(req, v1rset),
                          [('managepermission', actions.ManagePermissionsAction),
                           ('addrelated', actions.AddRelatedActions),
                           ('revhead', vcsactions.VCHeadRevisionAction),
                           ('revall', vcsactions.VCAllRevisionsAction),
#                           ('revnext', vcsactions.VCNextRevisionAction),
                           ('revfiles', vcsactions.VCRevisionModifiedFilesAction),
                           ('revallfiles', vcsactions.VCRevisionAllFilesAction)])
        v2rset = req.execute('Any X WHERE X from_revision R, R revision 2, X content_for VF, VF name "toto.txt"')
        self.assertEqual(self.pactions(req, v2rset),
                          [('managepermission', actions.ManagePermissionsAction),
                           ('addrelated', actions.AddRelatedActions),
                           ('revall', vcsactions.VCAllRevisionsAction),
#                           ('revprev', vcsactions.VCPreviousRevisionAction),
                           ('revfiles', vcsactions.VCRevisionModifiedFilesAction),
                           ('revallfiles', vcsactions.VCRevisionAllFilesAction)])
        v3rset = req.execute('Any X WHERE X from_revision R, R revision 3, X content_for VF, VF name "data.bin.gz"')
        self.assertEqual(self.pactions(req, v2rset),
                          [('managepermission', actions.ManagePermissionsAction),
                           ('addrelated', actions.AddRelatedActions),
                           ('revall', vcsactions.VCAllRevisionsAction),
#                           ('revprev', vcsactions.VCPreviousRevisionAction),
                           ('revfiles', vcsactions.VCRevisionModifiedFilesAction),
                           ('revallfiles', vcsactions.VCRevisionAllFilesAction)])

    def test_metadata_view(self):
        v1 = self.execute('Any X WHERE X from_revision R, R revision 1, X content_for VF, VF name "toto.txt"').get_entity(0, 0)
        self.assertMultiLineEqual(remove_eid(v1.view('metadata')),
                              '<div class="metadata">#EID - <span>revision 1 of</span> <span class="value">toto.txt</span>,&nbsp;<span>created on</span> <span class="value">2007/12/19 11:18</span>&nbsp;<span>by</span> <span class="value">syt</span></div>')
        v3 = self.execute('Any X WHERE X from_revision R, R revision 3').get_entity(0, 0)
        self.assertMultiLineEqual(remove_eid(v3.view('metadata')),
                          '<div class="metadata">#EID - <span>revision 3 of</span> <span class="value">dir1/data.bin.gz</span>,&nbsp;<span>created on</span> <span class="value">2008/01/08 15:56</span>&nbsp;<span>by</span> <span class="value">syt</span></div>')


    def test_primary_view(self):
        req = self.request()
        v1rset = req.execute('Any X WHERE X from_revision R, R revision 1, X content_for VF, VF name "toto.txt"')
        view = self.vreg['views'].select('primary', req, rset=v1rset)
        self.assertIsInstance(view, svnfviews.VCPrimaryView)
        view.render() # just check no error raised
        v3rset = req.execute('Any X WHERE X from_revision R, R revision 3')
        view = self.vreg['views'].select('primary', req, rset=v3rset)
        self.assertIsInstance(view, svnfviews.DVCPrimaryView)
        view.render() # just check no error raised


class VFViewsTC(VCSFileTC):

    def test_entity(self):
        toto = self.execute('VersionedFile X WHERE X name %(name)s', {'name': 'toto.txt'}).get_entity(0, 0)
        self.assertIsInstance(toto, entities.VersionedFile)
        self.assertEqual(toto.name, "toto.txt")
        self.assertEqual(toto.directory, "")
        self.assertEqual(toto.path, "toto.txt")
        self.assertEqual(toto.head.revision, 2)
        self.assertTrue(toto.repository.path.endswith('testrepo'))
        tutu = self.execute('VersionedFile X WHERE X name %(name)s', {'name': 'tutu.txt'}).get_entity(0, 0)
        self.assertIsInstance(tutu, entities.VersionedFile)
        self.assertEqual(tutu.name, "tutu.txt")
        self.assertEqual(tutu.directory, "dir1")
        self.assertEqual(tutu.path, "dir1/tutu.txt")
        self.assertEqual(tutu.head.revision, 4)
        self.assertTrue(tutu.repository.path.endswith('testrepo'))
        self.assertEqual(tutu.deleted_in_branch(), False)
        data = self.execute('VersionedFile X WHERE X name %(name)s', {'name': 'data.bin.gz'}).get_entity(0, 0)
        self.assertEqual(data.head.revision, 3)
        self.assertEqual(data.deleted_in_branch(), True)

    def test_possible_actions(self):
        req = self.request()
        totorset = req.execute('VersionedFile X WHERE X name %(name)s', {'name': 'toto.txt'})
        self.assertEqual(self.pactions(req, totorset),
                          # XXX because we've some editable relation
                          #     check we actually want this
                          [('edit', actions.ModifyAction),
                           ('managepermission', actions.ManagePermissionsAction),
                           ('addrelated', actions.AddRelatedActions),
                           ('vfnewrevaction', vcsactions.VFPutUpdateAction),
                           ('vcsrmaction', vcsactions.VFRmAction),
                           ('revhead', vcsactions.VFHEADRevisionAction),
                           ('revall', vcsactions.VFAllRevisionsAction),
                           ('addrevision', vcsactions.VFAddRevisionAction)])


    def test_primary_view(self):
        req = self.request()
        totorset = req.execute('VersionedFile X WHERE X name %(name)s', {'name': 'toto.txt'})
        view = self.vreg['views'].select('primary', req, rset=totorset)
        self.assertIsInstance(view, primary.PrimaryView)
        view.render() # just check no error raised
        datarset = req.execute('VersionedFile X WHERE X name %(name)s', {'name': 'data.bin.gz'})
        view = self.vreg['views'].select('primary', req, rset=datarset)
        self.assertIsInstance(view, primary.PrimaryView)
        view.render() # just check no error raised


class UtilsTC(VCSFileTC):

    def test_rql_revision_content(self):
        rset = self.execute(*entities.rql_revision_content(self.repoeid, 1))
        self.assertEqual([(x.file.path, x.revision) for x in rset.entities()],
                          [('toto.txt', 1), ('dir1/data.bin.gz', 1)])
        rset = self.execute(*entities.rql_revision_content(self.repoeid, 2))
        self.assertEqual([(x.file.path, x.revision) for x in rset.entities()],
                          [('toto.txt', 2), ('dir1/data.bin.gz', 1)])
        rset = self.execute(*entities.rql_revision_content(self.repoeid, 3))
        self.assertEqual([(x.file.path, x.revision) for x in rset.entities()],
                          [('toto.txt', 2), ('dir1/data.bin.gz', 3)])
        rset = self.execute(*entities.rql_revision_content(self.repoeid, 4))
        self.assertEqual([(x.file.path, x.revision) for x in rset.entities()],
                          [('toto.txt', 2), ('dir1/tutu.txt', 4)])


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



Binary = repo.system_source._binary

for vceid, repoeid, dir, name, revnum in rql(
    'Any VC,REPO,D,N,R WHERE VC is VersionContent, '
    'VC from_revision REV, REV from_repository REPO, REV revision R, '
    'VC content_for VF, VF directory D, VF name N'):
    key = (str(repoeid), dir.encode('utf-8'), name.encode('utf-8'), str(revnum))
    key = '\x01'.join(key)
    sql('UPDATE cw_VersionContent as VC '
        'SET cw_data=%(data)s WHERE cw_eid=%(eid)s',
        {'data': Binary(key), 'eid': vceid}, ask_confirm=False)

sync_schema_props_perms('VersionContent', syncperms=False)

sql('DROP FUNCTION IF EXISTS version_data(bigint)')
sql('DROP FUNCTION IF EXISTS sys_path_init(text, text)')



"""some facets to filter vcsfile entities

:organization: Logilab
:copyright: 2008-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
:license: General Public License version 2 - http://www.gnu.org/licenses
"""
__docformat__ = "restructuredtext en"

from cubicweb.selectors import is_instance
from cubicweb.web import facet

class RepositoryTypeFacet(facet.AttributeFacet):
    __regid__ = 'vcsfile.repo.type'
    __select__ = facet.AttributeFacet.__select__ & is_instance('Repository')
    rtype = 'type'

class RevisionBranchFacet(facet.AttributeFacet):
    __regid__ = 'vcsfile.rev.branch'
    __select__ = facet.AttributeFacet.__select__ & is_instance('Revision')
    rtype = 'branch'
    i18nable = False

class RevisionCreationDateFacet(facet.DateRangeFacet):
    __regid__ = 'vcsfile.rev.cd'
    __select__ = facet.DateRangeFacet.__select__ & is_instance('Revision')
    rtype = 'creation_date'
    order = 4

class VFModificationDateFacet(facet.DateRangeFacet):
    __regid__ = 'vcsfile.vf.cd'
    __select__ = facet.DateRangeFacet.__select__ & is_instance('VersionedFile')
    rtype = 'modification_date'
    order = 4



"""cubicweb-vcsfile"""

IMMUTABLE_ATTRIBUTES = frozenset(('VersionedFile.directory',
                                  'VersionedFile.name',
                                  'VersionedFile.from_repository',
                                  'Revision.revision',
                                  'Revision.changeset',
                                  'Revision.branch',
                                  'Revision.from_repository',
                                  'DeletedVersionContent.from_revision',
                                  'DeletedVersionContent.content_for',
                                  'VersionContent.from_revision',
                                  'VersionContent.content_for',
                                  'VersionContent.data',
                                  ))



"""utils for vcsfile

:organization: Logilab
:copyright: 2007-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

import os
import os.path as osp

from mercurial import ui, commands as hgcmd



from yams.buildobjs import RelationDefinition

class tags(RelationDefinition):
    subject = 'Tag'
    object = 'VersionedFile'

class filed_under(RelationDefinition):
    subject = 'VersionedFile'
    object = 'Folder'        



from os.path import join, abspath

from logilab.common.decorators import clear_cache

from cubicweb import ValidationError
from cubicweb.devtools.testlib import TestCase, CubicWebTC

from cubes.vcsfile.entities import remove_authinfo
from utils import VCSFileTC, HERE


class RemoveAuthInfoTC(TestCase):
    def test(self):
        self.assertEqual(remove_authinfo('http://www.logilab.org/src/cw'),
                          'http://www.logilab.org/src/cw')
        self.assertEqual(remove_authinfo('http://toto:1223@www.logilab.org/src/cw'),
                          'http://***@www.logilab.org/src/cw')


class BranchHeadsTC(VCSFileTC):
    _repo_path = u'testrepohg_branches'
    repo_type = u'mercurial'

    def rev(self, file, *args):
        return self.vcsrepo.versioned_file('', file).branch_head(*args).revision

    def test_vcsrm(self):
        self.grant_write_perm('managers')
        # XXX outcome depends on method names (=> execution order)
        vf = self.vcsrepo.versioned_file('', 'README')
        vf.vcs_rm()
        self.commit()
        clear_cache(vf, 'branch_head')
        clear_cache(vf, 'version_content')
        self.assertTrue(vf.deleted_in_branch())
        self.assertTrue(vf.head.rev.parent_revision)
        self.assertEqual(len(vf.head.rev.reverse_at_revision), 3,
                          vf.head.rev.reverse_at_revision)

    def test_branch_head_no_arg(self):
        self.assertEqual(self.rev('README'), 1)
        self.assertEqual(self.rev('file.txt'), 4)
        self.assertEqual(self.rev('otherfile.txt'), 4)

    def test_branch_head_default(self):
        self.assertEqual(self.rev('README', 'default'), 1)
        self.assertEqual(self.rev('file.txt', 'default'), 4)
        self.assertEqual(self.rev('otherfile.txt', 'default'), 4)

    def test_branch_head_branch1(self):
        self.assertEqual(self.rev('README', 'branch1'), 1)
        self.assertEqual(self.rev('file.txt', 'branch1'), 2)
        self.assertEqual(self.rev('otherfile.txt', 'branch1'), 3)


class RepositoryTC(CubicWebTC):
    def test_repo_path_security(self):
        req = self.request()
        path = unicode(abspath(join(HERE, 'whatever')))
        vcsrepo = req.create_entity('Repository',
                                    path=path, type=u'mercurial')
        req.cnx.commit()
        self.assertEqual(vcsrepo.dc_title(),
                          u'mercurial:%s' % path)
        self.login('anon')
        req = self.request()
        vcsrepo = req.execute('Repository X').get_entity(0, 0)
        self.assertEqual(vcsrepo.dc_title(), 'mercurial repository #%s' % vcsrepo.eid)

    def test_unmodifiable_attrs(self):
        req = self.request()
        self.assertRaises(ValidationError, req.create_entity, 'Repository',
                          source_url=u"http://myrepo.ru",
                          type=u'mercurial')
        req.cnx.rollback()
        req = self.request()
        repo = req.create_entity('Repository', type=u'mercurial',
                                 path=unicode(abspath(join(HERE, 'whatever'))))
        req.cnx.commit()
        self.assertRaises(ValidationError, repo.set_attributes, path=u'new')
        self.assertRaises(ValidationError, repo.set_attributes, type=u'mercurial')
        self.assertRaises(ValidationError, repo.set_attributes, source_url=u'http://myrepo.fr')

    def test_source_url_format(self):
        req = self.request()
        self.assertRaises(ValidationError, req.create_entity, 'Repository',
                          path=u'whatever', type=u'mercurial')
        self.assertRaises(ValidationError, req.create_entity, 'Repository',
                          source_url=u'whatever', type=u'mercurial')


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



from utils import VCSFileTC

class HooksTC(VCSFileTC):

    def setup_database(self):
        self.request().create_entity('Tag', name=u'dir1')
        self.request().create_entity('Folder', name=u'dir1')

    def test_auto_classification(self):
        toto = self.execute('Any X WHERE X name "toto.txt"').get_entity(0, 0)
        self.failIf(toto.reverse_tags)
        self.failIf(toto.filed_under)
        tutu = self.execute('Any X WHERE X name "tutu.txt"').get_entity(0, 0)
        self.failUnless(tutu.reverse_tags)
        self.failUnless(tutu.filed_under)


class AtRevisionHooksTC(VCSFileTC):
    _repo_path = u'testrepohg_branches'
    repo_type = u'mercurial'

    def test_at_revision_is_set(self):
        self.failUnless(self.execute('Any X WHERE X at_revision Y'))
        def files(revnum):
            rev = self.execute('Revision X WHERE X revision %(rev)s', {'rev': revnum}).get_entity(0, 0)
            return sorted((x.file.name, x.revision) for x in rev.reverse_at_revision)
        self.assertEqual(files(0), [(u'README', 0), (u'file.txt', 0)])
        self.assertEqual(files(1), [(u'README', 1), (u'file.txt', 0)])
        self.assertEqual(files(2), [(u'README', 1), (u'file.txt', 2)])
        self.assertEqual(files(3), [(u'README', 1), (u'file.txt', 2), (u'otherfile.txt', 3)])
        self.assertEqual(files(4), [(u'README', 1), (u'file.txt', 4), (u'otherfile.txt', 4)])

    def test_vf_creation_date(self):
        vf = self.vcsrepo.versioned_file('', 'file.txt')
        self.failIf(vf.creation_date == vf.modification_date)
        rev = vf.branch_head().rev
        self.assertEqual(vf.modification_date, rev.creation_date)

if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()




# -*- coding: utf-8 -*-
import os.path as osp
import logging
logging.basicConfig(level=logging.CRITICAL)

from logilab.common.testlib import unittest_main
from cubicweb import ValidationError, devtools # import devtools to ensure cubes are importable
from cubes.vcsfile import bridge

from utils import VCSFileTC

class SVNRepoTC(VCSFileTC):
    repo_type = u'subversion'
    _repo_path = u'testrepo'
    repo_encoding = u'utf-8'

    def test_base(self):
        res = self.execute('Revision X')
        self.assertEqual(len(res), 7)
        res = self.execute('VersionedFile X')
        self.assertEqual(len(res), 5)
        res = self.execute('VersionContent X')
        self.assertEqual(len(res), 6)
        res = self.execute('DeletedVersionContent X')
        self.assertEqual(len(res), 2)

    def _test_advanced_rev_query(self, res):
        self.assertEqual(res[0], [1, 'revision 1', '2007-12-19 11:18', u'syt', None, None])
        self.assertEqual(res[1], [2, 'revision 2: modif toto.txt', '2007-12-19 11:18', u'syt', None, None])
        self.assertEqual(res[2], [3, 'delete binary file', '2008-01-08 15:56', u'syt', None, None])
        self.assertEqual(res[3], [4, 'hop', '2008-01-09 18:11', u'syt', None, None])

    def _test_advanced_vc_query(self, res):
        self.assertEqual(res[0], ['__hg_needs_a_file__', 5, None, u'application/octet-stream',
                                   'the file hg needs\n'])
        self.assertEqual(res[1], ['data.bin.gz', 1, u'gzip', u'application/octet-stream',
                                   '\x1f\x8b\x08\x00\xde\x17YG\x00\x03\x0b\xc9\xc8,V\x00\xa2D\x85\x92\xd4\xe2\x12.\x00\x8c-\xc0\xfa\x0f\x00\x00\x00'])
        self.assertEqual(res[2], [u'm\xe9', 7, None, u'application/octet-stream', ''])
        self.assertEqual(res[3], ['toto.txt', 1, self.repo_encoding, u'text/plain',
                                   'hop\n'])
        self.assertEqual(res[4], ['toto.txt', 2, self.repo_encoding, u'text/plain',
                                   'hop\nhop\n'])
        self.assertEqual(res[5], ['tutu.txt', 4, self.repo_encoding, u'text/plain',
                                   'coucou\n'])

    def _test_advanced_dvc_query(self, res):
        self.assertEqual(res[0], [u'__hg_needs_a_file__', 6])
        self.assertEqual(res[1], [u'data.bin.gz', 3])

    def test_advanced(self):
        res = self.execute('Any X,D,N ORDERBY D,N WHERE X is VersionedFile, X directory D, X name N')
        self.assertEqual(len(res), 5)
        self.assertEqual(res[0][1:], [u'', u'm\xe9'])
        self.assertEqual(res[1][1:], [u'', u'toto.txt'])
        self.assertEqual(res[2][1:], [u'dir1', 'data.bin.gz'])
        self.assertEqual(res[3][1:], [u'dir1', 'tutu.txt'])
        self.assertEqual(res[4][1:], [u'dir1/ghostdir', '__hg_needs_a_file__'])
        res = self.execute('Any R,C,D,A,CS,B ORDERBY R '
                           'WHERE X revision R, X description C, '
                           'X creation_date D, X author A, X changeset CS, X branch B')
        for row in res:
            row[2] = row[2].strftime('%Y-%m-%d %H:%M')
        self.assertEqual(len(res), 7)
        self._test_advanced_rev_query(res)
        res = self.execute('Any P,R,E,F,DA ORDERBY P,R '
                           'WHERE X content_for Y, Y name P, X from_revision REV, '
                           'REV revision R, X data_encoding E, X data_format F, '
                           'X data DA')
        for row in res:
            row[-1] = row[-1].getvalue()
        self.assertEqual(len(res), 6)
        self._test_advanced_vc_query(res)
        res = self.execute('Any P,R ORDERBY P,R '
                           'WHERE X content_for Y, Y name P, X from_revision REV, '
                           'REV revision R, X is DeletedVersionContent')
        self.assertEqual(len(res), 2)
        self._test_advanced_dvc_query(res)

    def test_revision_parent(self):
        res = self.execute('Any R,P ORDERBY REV WHERE R revision REV, R parent_revision P?')
        self.assertEqual(len(res), 7)
        self.assertEqual(res[0][1], None)
        latestrev = res[0][0]
        for i, (rev, parent) in enumerate(res[1:]):
            self.assertEqual(parent, latestrev)
            latestrev = rev

    first_revision = 1
    def test_revision_search(self):
        res = self.execute('Any P ORDERBY P WHERE X from_revision R, R revision %s, '
                           'X content_for VF, VF name P' % self.first_revision)
        self.assertEqual(len(res), 2)
        self.assertEqual(res[0], [u'data.bin.gz'])
        self.assertEqual(res[1], [u'toto.txt'])

    author = 'syt'
    def test_author_search(self):
        res = self.execute('DISTINCT Any P ORDERBY P WHERE X from_revision R, '
                           'R author %(author)s, X content_for VF, VF name P',
                           {'author': self.author})
        self.assertEqual(len(res), 3)
        self.assertEqual(res[0], [u'data.bin.gz'])
        self.assertEqual(res[1], [u'toto.txt'])
        self.assertEqual(res[2], [u'tutu.txt'])

    def test_anon_access_data(self):
        bridge._REPOHDLRS.clear()
        self.login('anon')
        res = self.execute('VersionContent X LIMIT 1').get_entity(0, 0)
        self.failUnless(res.data)
        self.failUnless(res.data.getvalue())

class SVNPartRepoTC(VCSFileTC):
    repo_type = u'subversion'
    _repo_path = u'testrepo'
    repo_encoding = u'latin1'
    repo_subpath = u'dir1'

    def test_base(self):
        res = self.execute('Revision X')
        self.assertEqual(len(res), 7)
        res = self.execute('VersionedFile X')
        self.assertEqual(len(res), 3)
        res = self.execute('VersionContent X')
        self.assertEqual(len(res), 3)
        res = self.execute('DeletedVersionContent X')
        self.assertEqual(len(res), 2)

class SVNNoContentPartRepoTC(VCSFileTC):
    repo_type = u'subversion'
    _repo_path = u'testrepo'
    repo_encoding = u'latin1'
    repo_subpath = u'dir1'

    def setUp(self):
        # call parent of parent setUp
        super(VCSFileTC, self).setUp()
        self.repo.config.global_set_option('import-revision-content', False)
        VCSFileTC.setUp(self, False)

    def test_base(self):
        res = self.execute('Revision X')
        self.assertEqual(len(res), 7)
        res = self.execute('VersionedFile X')
        self.assertEqual(len(res), 0)
        res = self.execute('VersionContent X')
        self.assertEqual(len(res), 0)
        res = self.execute('DeletedVersionContent X')
        self.assertEqual(len(res), 0)


class HGRepoTC(SVNRepoTC):
    repo_type = u'mercurial'
    _repo_path = u'testrepohg'
    repo_encoding = u'latin1'

    first_revision = 0
    author = 'Sylvain Thenault <sylvain.thenault@logilab.fr>'

    def _test_advanced_rev_query(self, res):
        self.assertEqual(res[0], [0, 'revision 1', '2008-08-19 13:37',
                                   u'Sylvain Thenault <sylvain.thenault@logilab.fr>',  u'1a4a39caf4fe', 'default'])
        self.assertEqual(res[1], [1, 'revision 2: modif toto.txt', '2008-08-19 13:37',
                                   u'Sylvain Thenault <sylvain.thenault@logilab.fr>', u'295b01eaded6', 'default'])
        self.assertEqual(res[2], [2, u'delete binary file', '2008-08-19 13:38',
                                   u'Sylvain Thenault <sylvain.thenault@logilab.fr>', u'9b1d7e2c028e', 'default'])
        self.assertEqual(res[3], [3, 'hop', '2008-08-19 13:38',
                                   u'Sylvain Thenault <sylvain.thenault@logilab.fr>', u'45e1887b0167', 'default'])

    def _test_advanced_vc_query(self, res):
        self.assertEqual(res[1], ['data.bin.gz', 0, u'gzip', u'application/octet-stream',
                                   '\x1f\x8b\x08\x08\xed\x93\xaaH\x00\x03data.bin\x00\x93\xef\xe6\xe0\xd81y\x95\x07\x03sI~I\xbe^IE\t\xc3\xe9\x13\xfa\x0f\x1e110\xb8\xbesbe``\x00\x00\x85\xb3]\x01"\x00\x00\x00'])
        self.assertEqual(res[2], [u'm\xe9', 6, None, u'application/octet-stream', ''])
        self.assertEqual(res[3], ['toto.txt', 0, u'latin1', u'text/plain',
                                   'hop\n\n'])
        self.assertEqual(res[4], ['toto.txt', 1, u'latin1', u'text/plain',
                                   'hop\nhop\n\n'])
        self.assertEqual(res[5], ['tutu.txt', 3, u'latin1', u'text/plain',
                                   'coucou\n\n'])

    def _test_advanced_dvc_query(self, res):
        self.assertEqual(res[1], [u'data.bin.gz', 2])


    def test_error_update_repo(self):
        self.assertRaises(ValidationError,
                          self.execute, 'SET X type "subversion" WHERE X is Repository')


class HGPartRepoTC(SVNPartRepoTC):
    repo_type = u'mercurial'
    _repo_path = u'testrepohg'
    repo_encoding = u'latin1'

class HGNoContentPartRepoTC(SVNNoContentPartRepoTC):
    repo_type = u'mercurial'
    _repo_path = u'testrepohg'
    repo_encoding = u'latin1'


class HGCopyAndRenameTC(VCSFileTC):
    _repo_path = u'testrepohg_renaming'
    repo_type = u'mercurial'

    def test_copy(self):
        rset = self.execute('Any F1, R1, F2, R2 WHERE VC1 vc_copy VC2,'
                            'VC1 content_for VF1, VF1 name F1, VC1 from_revision REV1, REV1 revision R1,'
                            'VC2 content_for VF2, VF2 name F2, VC2 from_revision REV2, REV2 revision R2'
                            )
        self.assertEqual(rset.rows,
                          [[u'tete.txt', 3, u'tata.txt', 2]])

    def test_rename(self):
        rset = self.execute('Any F1, R1, F2, R2 WHERE VC1 vc_rename VC2,'
                            'VC1 content_for VF1, VF1 name F1, VC1 from_revision REV1, REV1 revision R1,'
                            'VC2 content_for VF2, VF2 name F2, VC2 from_revision REV2, REV2 revision R2'
                            )
        self.assertEqual(rset.rows,
                          [[u'tata.txt', 2, u'toto.txt', 0]])

class HGCachedRepoTC(VCSFileTC):
    repo_type = u'mercurial'
    _repo_path = u'testrepohg'
    repo_encoding = u'latin1'

    def _create_repo(self):
        return self.session.create_entity('Repository', type=self.repo_type,
                                          source_url='file://' + self.repo_path,
                                          encoding=self.repo_encoding)

    def test_cache_mgmt(self):
        self.session.repo = self.repo
        self.failIf(osp.isabs(self.vcsrepo.local_cache))
        cachedir = osp.join(self.config['local-repo-cache-root'],
                            self.vcsrepo.local_cache)
        self.failUnless(osp.isabs(cachedir))
        self.failUnless(osp.isdir(cachedir))
        self.execute('DELETE Repository R WHERE R eid %(r)s', {'r': self.vcsrepo.eid})
        self.commit()
        self.failIf(osp.isdir(cachedir))

if __name__ == '__main__':
    unittest_main()




def new_revision_rql(has_parent, set_all=False):
    """return rql to insert a revision entity

    substitutions used:
    * description, author, branch, repoeid (mandatory)
    * parent (if `has_parent` is True)
    * date, revision, changeset (if `set_all` is True)
    """
    rql = ('INSERT Revision R: R description %(description)s, '
           'R author %(author)s, R branch %(branch)s, R from_repository X')
    if set_all:
        rql += ', R revision %(revision)s, R changeset %(changeset)s'
        rql += ', R creation_date %(date)s' # XXX use specific field
    if has_parent:
        rql += ', R parent_revision PR'
    rql += ' WHERE X eid %(repoeid)s'
    if has_parent:
        rql += ', PR eid %(parent)s'
    return rql

def new_file_rql():
    return ('INSERT VersionedFile X: X from_repository R, '
            'X directory %(directory)s, X name %(name)s WHERE R eid %(repoeid)s')

def new_file_content_rql():
    return ('INSERT VersionContent X: X content_for VF, X from_revision REV, '
            'X data %(data)s WHERE VF eid %(vfeid)s, REV eid %(reveid)s')



# -*- coding: utf-8 -*-
from shutil import rmtree, copytree
from os import system, unlink, chdir, getcwd
from os.path import join, split, exists, abspath
import tempfile

from cubicweb import Binary, QueryError, ValidationError, Unauthorized
from cubicweb import devtools # makes cubes importable
from cubes.vcsfile import entities, bridge

from utils import VCSFileTC

REPOBASEPATH = abspath(split(__file__)[0])

class _WriteTC(VCSFileTC):
    _repo_path = u'testrepocopy'
    def setUp(self):
        try:
            copytree(str(self.orig_repo_path), str(self.repo_path))
            VCSFileTC.setUp(self)
            self.copath = join(REPOBASEPATH, self._repo_path)
        except:
            rmtree(str(self.repo_path))
            raise
        self.grant_write_perm('managers')

    def tearDown(self):
        VCSFileTC.tearDown(self)
        rmtree(str(self._repo_path))

    def _test_new_revision(self, entity, vf, rev, revision):
        self.assertEqual(entity.content_for[0].eid, vf.eid)
        self.assertEqual(entity.description, u'add hôp')
        self.assertEqual(entity.author, 'syt')
        self.assertEqual(entity.data.getvalue(), 'hop\nhop\nhop\n')
        self.assertEqual(entity.data_encoding.lower(), self.repo_encoding)
        self.assertEqual(entity.data_format, 'text/plain')
        self.assertEqual(entity.rev.eid, rev.eid)
        self.assertEqual(entity.rev.revision, revision)
        self.assertTrue(entity.rev.parent_revision)
        self.assertTrue(entity.rev.reverse_at_revision)

    def _new_toto_revision(self, data='hop\nhop\nhop\n', branch=entities._MARKER):
        vf = self.execute('VersionedFile X WHERE X name "toto.txt"').get_entity(0, 0)
        rev = self.vcsrepo.make_revision(msg=u'add hôp', author=u'syt', branch=branch)
        vc = self.execute('INSERT VersionContent X: X content_for VF, X from_revision R, '
                           'X data %(data)s WHERE VF eid %(vf)s, R eid %(r)s',
                           {'vf': vf.eid, 'r': rev.eid,
                            'data': Binary(data)}).get_entity(0, 0)
        return vf, rev, vc

    def test_new_revision(self):
        vf, rev, vc = self._new_toto_revision()
        self.commit()
        self._test_new_revision(vc, vf, rev, 8 + self.repo_rev_offset)
        try:
            session = self.repo.internal_session()
            bridge.import_content(session)
        finally:
            session.close()
        self.assertEqual(len(self.execute('Any X WHERE X revision %(r)s', {'r': rev.revision})), 1)

    def test_new_revision_security(self):
        self.execute('DELETE CWPermission X')
        self.commit()
        self.assertRaises(Unauthorized, self._new_toto_revision)

# XXX unable to do this, we need to get the associated VersionedFile in source.add_entity...
#     def test_two_steps_new_revision(self):
#         vf = self.execute('VersionedFile X WHERE X name "toto.txt"')
#         eid = self.execute('INSERT VersionContent X: X description %(msg)s, X author %(author)s, '
#                            'X data %(data)s', {'msg': u'add hôp', 'author': u'syt',
#                                                'data': Binary('hop\nhop\nhop\n')})[0][0]
#         self.execute('SET X content_for VF WHERE X eid %(x)s, VF eid %(vf)s',
#                      {'vf': vf[0][0], 'x': eid}, ('vf', 'x'))
#         self.commit()
#         self._test_new_revision(vf, eid)

    def test_rollback_new_revision(self):
        vf, rev, vc = self._new_toto_revision()
        self.rollback()
        # fail due to a bug in pysqlite, see
        # http://oss.itsystementwicklung.de/trac/pysqlite/ticket/219
        # http://www.sqlite.org/cvstrac/tktview?tn=2765,3
        # http://www.initd.org/pub/software/pysqlite/doc/usage-guide.html#controlling-transactions
        #rset = self.execute('Any X WHERE X eid %(x)s', {'x': rev.eid})
        #self.failIf(rset)
        rset = self.execute('Any X WHERE X eid %(x)s', {'x': vc.eid})
        self.failIf(rset)
        rset = self.execute('VersionContent X WHERE X content_for VF, VF eid %(vf)s', {'vf': vf.eid})
        self.assertEqual(len(rset), 2)
        # XXX check actually not in the repository

    def test_error_new_revision(self):
        # can't specify revision number
        self.assertRaises(QueryError, self.execute,
                          'INSERT Revision X: X from_repository R, '
                           'X description %(msg)s, X author %(author)s, X revision 2 '
                           'WHERE R eid %(r)s',
                           {'r': self.repoeid, 'msg': u'add hôp', 'author': u'syt'})
        # missing from_repository
        self.assertRaises(ValidationError, self.execute,
                          'INSERT Revision X: '
                           'X description %(msg)s, X author %(author)s '
                           'WHERE R eid %(r)s',
                           {'r': self.repoeid, 'msg': u'add hôp', 'author': u'syt'})
        self.rollback()
        # OK
        self.execute('INSERT Revision X: X from_repository R, '
                     'X description %(msg)s, X author %(author)s '
                     'WHERE R eid %(r)s',
                     {'r': self.repoeid, 'msg': u'add hôp', 'author': u'syt'})
        # try to create another revision
        self.assertRaises(QueryError, self.execute,
                          'INSERT Revision X: X from_repository R, '
                          'X description %(msg)s, X author %(author)s '
                          'WHERE R eid %(r)s',
                          {'r': self.repoeid, 'msg': u'add hôp', 'author': u'syt'})
        # commit while nothing changed in the revision
        self.assertRaises(QueryError, self.commit)

    def test_error_new_version_content(self):
        vfeid = self.execute('VersionedFile X WHERE X name "toto.txt"')[0][0]
        badreveid = self.execute('Revision X WHERE X revision 1')[0][0]
        # no revision transaction
        self.assertRaises(QueryError, self.execute,
                          'INSERT VersionContent X: X content_for VF, X from_revision R, '
                           'X data %(data)s WHERE VF eid %(vf)s, R eid %(r)s',
                          {'vf': vfeid, 'r': badreveid,
                           'data': Binary('hop\nhop\nhop\n')})
        # missing content_for relation
        reveid = self.execute('INSERT Revision X: X from_repository R, '
                              'X description %(msg)s, X author %(author)s '
                              'WHERE R eid %(r)s',
                              {'r': self.repoeid, 'msg': u'add hôp', 'author': u'syt'})[0][0]
        self.assertRaises(ValidationError, self.execute,
                          'INSERT VersionContent X: X from_revision R, '
                           'X data %(data)s WHERE R eid %(r)s',
                          {'r': reveid,
                           'data': Binary('hop\nhop\nhop\n')})
        self.rollback()
        # missing data attribute
        self.assertRaises(ValidationError, self.execute,
                          'INSERT VersionContent X: X content_for VF, X from_revision R '
                           'WHERE VF eid %(vf)s, R eid %(r)s',
                          {'vf': vfeid, 'r': badreveid})
        self.rollback()
        # missing from_revision relation
        reveid = self.execute('INSERT Revision X: X from_repository R, '
                              'X description %(msg)s, X author %(author)s '
                              'WHERE R eid %(r)s',
                              {'r': self.repoeid, 'msg': u'add hôp', 'author': u'syt'})[0][0]
        self.execute('INSERT VersionContent X: X content_for VF, '
                     'X data %(data)s WHERE VF eid %(vf)s',
                     {'vf': vfeid, 'data': Binary('hop\nhop\nhop\n')})
        self.assertRaises(ValidationError, self.commit)
        self.rollback()
        # bad from_revision relation
        reveid = self.execute('INSERT Revision X: X from_repository R, '
                              'X description %(msg)s, X author %(author)s '
                              'WHERE R eid %(r)s',
                              {'r': self.repoeid, 'msg': u'add hôp', 'author': u'syt'})[0][0]
        self.execute('INSERT VersionContent X: X content_for VF, X from_revision R, '
                     'X data %(data)s WHERE VF eid %(vf)s, R eid %(r)s',
                     {'vf': vfeid, 'r': badreveid,
                      'data': Binary('hop\nhop\nhop\n')})
        self.assertRaises(QueryError, self.commit)

    def test_new_file(self):
        vfeid = self.execute('INSERT VersionedFile X: X from_repository R, '
                          'X directory %(dir)s, X name %(name)s WHERE R eid %(repo)s',
                          {'repo': self.repoeid, 'dir': u'dir1', 'name': u'tutu.png'})[0][0]
        reveid = self.execute('INSERT Revision X: X from_repository R, '
                              'X description %(msg)s, X author %(author)s '
                              'WHERE R eid %(r)s',
                              {'r': self.repoeid, 'msg': u'new file', 'author': u'syt'})[0][0]
        eid = self.execute('INSERT VersionContent X: X content_for %(vf)s, X from_revision %(r)s, '
                           'X data %(data)s',
                           {'vf': vfeid, 'r': reveid, 'msg': u'new file', 'author': u'syt',
                            'data': Binary('zoubî')})[0][0]
        self.commit()
        entity = self.execute('Any X WHERE X eid %(x)s', {'x': eid}).get_entity(0, 0)
        vf = entity.content_for[0]
        self.assertEqual(vf.from_repository[0].eid, self.repoeid)
        self.assertEqual(vf.directory, 'dir1')
        self.assertEqual(vf.name, 'tutu.png')
        self.assertEqual(entity.description, 'new file')
        self.assertEqual(entity.author, 'syt')
        self.assertEqual(entity.data.getvalue(), 'zoubî')
        self.assertEqual(entity.data_encoding, None)
        self.assertEqual(entity.data_format, 'image/png')

    def test_error_new_file(self):
        self.execute('INSERT VersionedFile X: X directory %(dir)s, X name %(name)s WHERE R eid %(repo)s',
                     {'repo': self.repoeid, 'dir': u'dir1', 'name': u'tutu.txt'})
        self.assertRaises(ValidationError, self.commit)
        self.execute('INSERT VersionedFile X: X directory %(dir)s, X name %(name)s, X from_repository R '
                     'WHERE R eid %(repo)s', {'repo': self.repoeid, 'dir': u'dir1', 'name': u'tutu.txt'})
        self.assertRaises(QueryError, self.commit)

    def test_multiple_changes_and_deletion(self):
        reveid = self.execute('INSERT Revision X: X from_repository R, '
                              'X description %(msg)s, X author %(author)s '
                              'WHERE R eid %(r)s',
                              {'r': self.repoeid, 'msg': u'add hop', 'author': u'syt'})[0][0]
        vf1 = self.execute('VersionedFile X WHERE X name "toto.txt"')
        eid1 = self.execute('INSERT VersionContent X: X content_for %(vf)s, X from_revision %(r)s, '
                            'X data %(data)s',
                            {'vf': vf1[0][0], 'r': reveid,
                             'data': Binary('hop\nhop\nhop\n')})[0][0]
        vf2 = self.execute('INSERT VersionedFile X: X from_repository R, '
                           'X directory %(dir)s, X name %(name)s WHERE R eid %(repo)s',
                           {'repo': self.repoeid, 'dir': u'dir1', 'name': u'tutu.txt'})
        eid2 = self.execute('INSERT VersionContent X: X content_for %(vf)s, X from_revision %(r)s, '
                           'X data %(data)s',
                            {'vf': vf2[0][0], 'r': reveid,
                             'data':Binary('zoubî')})[0][0]
        self.commit()
        self.assertFalse(self.vcsrepo.is_directory_deleted('dir1'))
        entity1 = self.execute('Any X WHERE X eid %(x)s', {'x': eid1}).get_entity(0, 0)
        self.assertEqual(entity1.content_for[0].eid, vf1[0][0])
        self.assertEqual(entity1.description, 'add hop')
        self.assertEqual(entity1.author, 'syt')
        self.assertEqual(entity1.data.getvalue(), 'hop\nhop\nhop\n')
        entity2 = self.execute('Any X WHERE X eid %(x)s', {'x': eid2}).get_entity(0, 0)
        vf2 = entity2.content_for[0]
        self.assertEqual(vf2.from_repository[0].eid, self.repoeid)
        self.assertEqual(vf2.directory, 'dir1')
        self.assertEqual(vf2.name, 'tutu.txt')
        self.assertEqual(entity2.description, 'add hop')
        self.assertEqual(entity2.author, 'syt')
        self.assertEqual(entity2.data.getvalue(), 'zoubî')
        # deletion
        tmpdir = tempfile.mkdtemp()
        try:
            self.checkout(tmpdir)
            self.failUnless(exists(join(tmpdir, 'dir1', 'tutu.txt')))
            self.failUnless(exists(join(tmpdir, 'toto.txt')))
        finally:
            rmtree(tmpdir)
        reveid = self.execute('INSERT Revision X: X from_repository R, '
                              'X description %(msg)s, X author %(author)s '
                              'WHERE R eid %(r)s',
                              {'r': self.repoeid, 'msg': u'kill', 'author': u'auc'})[0][0]
        vf1 = self.execute('VersionedFile X WHERE X name "toto.txt"')
        # let's delete toto by linking its vf to DeletedVersionContent
        eid1 = self.execute('INSERT DeletedVersionContent X: X content_for %(vf)s, X from_revision %(r)s',
                           {'vf': vf1[0][0], 'r': reveid})[0][0]
        self.commit()
        entity1 = self.execute('Any X WHERE X eid %(x)s', {'x': eid1}).get_entity(0, 0)
        self.assertEqual(entity1.content_for[0].eid, vf1[0][0])
        self.assertEqual(entity1.description, 'kill')
        self.assertEqual(entity1.author, 'auc')
        tmpdir = tempfile.mkdtemp()
        try:
            self.checkout(tmpdir)
            self.failUnless(exists(join(tmpdir, 'dir1', 'tutu.txt')))
            self.failIf(exists(join(tmpdir, 'toto.txt')))
            self.failIf(exists(join(tmpdir, 'dir1', 'subdir')))
        finally:
            rmtree(tmpdir)
        oddfile = self.execute('VersionedFile F WHERE F name "__hg_needs_a_file__"').get_entity(0,0)
        content = vf1.get_entity(0,0).reverse_content_for
        self.assertEqual(len(content), 4)
        self.assertIsInstance(content[0], entities.DeletedVersionContent)
        content = oddfile.reverse_content_for
        self.assertEqual(len(content), 2)
        self.assertIsInstance(content[0], entities.DeletedVersionContent)
        self.failUnless(self.vcsrepo.is_directory_deleted('dir1/ghostdir'))


class SVNSourceWriteTC(_WriteTC):
    orig_repo_path = 'testrepo'
    repo_type = u'subversion'
    repo_rev_offset = 0

    def checkout(self, tmpdir):
        svncmd = 'svn co file://%s %s >/dev/null 2>&1' % (self.copath, tmpdir)
        if system(svncmd):
            raise Exception('can not check out %s' % self.copath)


class HGSourceWriteTC(_WriteTC):
    orig_repo_path = 'testrepohg'
    repo_type = u'mercurial'
    repo_encoding = u'latin1'
    repo_rev_offset = -1

    def checkout(self, tmpdir):
        svncmd = 'hg clone %s %s >/dev/null 2>&1' % (self.copath, join(tmpdir, 'testrepohg'))
        if system(svncmd):
            raise Exception('can not check out %s' % self.copath)
        system('mv %s %s' % (join(tmpdir, 'testrepohg', '*'), tmpdir))
        system('mv %s %s' % (join(tmpdir, 'testrepohg', '.hg'), tmpdir))

    def test_branches_from_app(self):
        vf, rev, vc = self._new_toto_revision(data='branch 1.2.3 content',
                                              branch=u'1.2.3')
        self.commit()
        hgdir = self.copath
        self.assertEqual(file(join(hgdir, 'toto.txt')).read(),
                          'hop\nhop\n\n')
        system('cd %s; hg up 1.2.3 >/dev/null 2>&1' % hgdir)
        self.assertEqual(file(join(hgdir, 'toto.txt')).read(),
                          'branch 1.2.3 content')
        # create new changeset in the default branch
        vf, rev, vc = self._new_toto_revision(data='branch default content')
        self.commit()
        system('cd %s; hg up default >/dev/null 2>&1' % hgdir)
        self.assertEqual(file(join(hgdir, 'toto.txt')).read(),
                          'branch default content')
        # create new changeset in the 1.2.3 branch
        vf, rev, vc = self._new_toto_revision(data='branch 1.2.3 new content',
                                              branch=u'1.2.3')
        self.commit()
        system('cd %s; hg up default >/dev/null 2>&1' % hgdir)
        self.assertEqual(file(join(hgdir, 'toto.txt')).read(),
                          'branch default content')
        # delete file from both branches
        vf.vcs_rm(branch=u'default', author=u'titi')
        self.commit()
        system('cd %s; hg up default >/dev/null 2>&1' % hgdir)
        self.failIf(exists((join(hgdir, 'toto.txt'))))
        vf.vcs_rm(branch=u'1.2.3', author=u'toto')
        self.commit()
        system('cd %s; hg up 1.2.3 >/dev/null 2>&1' % hgdir)
        self.failIf(exists((join(hgdir, 'toto.txt'))))
        vf.clear_all_caches()
        ex = self.assertRaises(QueryError, vf.vcs_rm, branch=u'1.2.3', author=u'toto')
        self.assertEqual(str(ex), 'toto.txt is already deleted from the vcs')

    def test_strip(self):
        self.assertEqual(self.vcsrepo.latest_known_revision(), 6)
        vcrset = self.execute('Any X WHERE X from_revision R, R revision 6')
        system('cd %s; hg --config extensions.hgext.mq= strip 6 2>/dev/null' % self.copath)
        try:
            session = self.repo.internal_session()
            bridge.import_content(session)
        finally:
            session.close()
        self.commit()
        self.assertEqual(self.vcsrepo.latest_known_revision(), 5)
        for eid, in vcrset:
            self.failIf(self.execute('Any X WHERE X eid %(x)s', {'x': eid}))

    def test_branches_from_hg(self):
        self.assertEqual(self.execute('Any COUNT(R) WHERE R is Revision').rows, [[7]])
        hgdir = self.copath
        system('cd %s; hg up default >/dev/null 2>&1' % hgdir)
        system('cd %s; hg branch newbranch >/dev/null 2>&1' % hgdir)
        system('cd %s; hg tag newbranch -m "opening newbranch"' % hgdir)
        try:
            session = self.repo.internal_session()
            bridge.import_content(session)
        finally:
            session.close()
        self.commit()
        hgtags = self.execute('VersionContent V WHERE V content_for VF,  '
                              'VF name ".hgtags"').get_entity(0,0)
        self.assertEqual(hgtags.from_revision[0].branch, u'newbranch')
        self.assertEqual(hgtags.from_revision[0].revision, 7)
        self.assertEqual(self.execute('Any COUNT(R) WHERE R is Revision').rows, [[8]])



if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



"template automatic tests"
import os
from logilab.common.testlib import unittest_main
from cubicweb.devtools import BaseApptestConfiguration, testlib
from utils import HERE, osp

class MyConfig(BaseApptestConfiguration):
    sourcefile = 'sources'

class AutomaticWebTest(testlib.AutomaticWebTest):
    configcls = MyConfig
    no_auto_populate = ('Repository', 'Revision', 'VersionedFile',
                        'VersionContent', 'DeletedVersionContent',)
    ignored_relations = set(('at_revision', 'parent_revision',
                             'from_repository', 'from_revision', 'content_for',))

    def custom_populate(self, how_many, cursor):
        self.request().create_entity('Repository', type=u'mercurial',
                                     path=unicode(osp.abspath(osp.join(HERE, 'testrepohg'))),
                                     encoding=u'latin1')

if __name__ == '__main__':
    unittest_main()
